# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
drvapi is not implemented in the simulator, but this module exists to allow
tests to import correctly.
"""
